# Copyright 2014 Sardar Yumatov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

if __name__ == '__main__':
    import os.path, sys
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))

import unittest, struct, pickle, json, datetime

# python 3 support
import six
import binascii

from apnsclient import Message, Session, APNs
from apnsclient.apns import Result
from apnsclient.backends.dummy import Backend as DummyBackend



class APNsClientMessageTest(unittest.TestCase):
    """ Test Message API. """

    def setUp(self):
        self.now = datetime.datetime.now()
        # Typical message
        self.message = Message(
            "0123456789ABCDEF",
            alert=u"Russian: \u0421\u0430\u0440\u0434\u0430\u0440",
            badge=10,
            sound="test.mp3",
            content_available=1,
            expiry=self.now + datetime.timedelta(days=3),
            priority=30,
            extra={'key': 'value'}
        )
        # Message with a custom payload
        self.raw_message = Message(
            ["0123456789ABCDEF", "FEDCBA9876543210"],
            payload=self.message.payload,
            priority=5,
            expiry=datetime.timedelta(days=5)
        )

    def test_payload(self):
        payload = self.message.get_json_payload()
        self.assertIsInstance(payload, six.binary_type)
        unicode_src = payload.decode('utf-8')
        payload = json.loads(unicode_src)
        self.assertEqual(payload["aps"], {
            "alert": self.message.alert,
            "badge": self.message.badge,
            "sound": self.message.sound,
            "content-available": self.message.content_available
        })
        for k, v in self.message.extra.items():
            self.assertEqual(payload[k], v)

    def test_serialization(self):
        # standard pickle
        s_message = pickle.dumps(self.message)
        s_raw_message = pickle.dumps(self.raw_message)
        c_message = pickle.loads(s_message)
        c_raw_message = pickle.loads(s_raw_message)

        for key in ('tokens', 'alert', 'badge', 'sound', 'content_available', 'expiry', 'extra', 'priority', '_payload'):
            self.assertEqual(getattr(self.message, key), getattr(c_message, key))
            self.assertEqual(getattr(self.raw_message, key), getattr(c_raw_message, key))

        # custom
        s_message = self.message.__getstate__()
        s_raw_message = self.raw_message.__getstate__()
        # JSON/XML/etc and store/send
        s_message = json.dumps(s_message)
        s_raw_message = json.dumps(s_raw_message)
        # unserialize
        s_message = json.loads(s_message)
        s_raw_message = json.loads(s_raw_message)
        # reconstruct
        c_message = Message(**s_message)
        c_raw_message = Message(**s_raw_message)

        for key in ('tokens', 'alert', 'badge', 'sound', 'content_available', 'expiry', 'extra', 'priority', '_payload'):
            self.assertEqual(getattr(self.message, key), getattr(c_message, key))
            self.assertEqual(getattr(self.raw_message, key), getattr(c_raw_message, key))

    def test_non_ascii(self):
        # meta-data size. ensure 'alert' is included.
        empty_msg_size = len(Message(tokens=[], alert="a").get_json_payload()) - 1

        MAX_UTF8_SIZE = 3  # size of maximum utf8 encoded character in bytes
        chinese_str = (
            u'\u5187\u869a\u5487\u6b8f\u5cca\u9f46\u9248\u6935\u4ef1\u752a'
            u'\u67cc\u521e\u62b0\u530a\u6748\u9692\u5c6e\u653d\u588f\u6678')
        chinese_msg_size = len(Message(tokens=[], alert=chinese_str).get_json_payload())
        self.assertLessEqual(
            chinese_msg_size,
            empty_msg_size + len(chinese_str) * MAX_UTF8_SIZE)

        MAX_EMOJI_SIZE = 4  # size of maximum utf8 encoded character in bytes
        # emoji
        emoji_str = (u'\U0001f601\U0001f603\U0001f638\U00002744')
        emoji_msg_size = len(Message(tokens="", alert=emoji_str).get_json_payload())
        self.assertLessEqual(
            emoji_msg_size,
            empty_msg_size + len(emoji_str) * MAX_EMOJI_SIZE)

    def test_batch(self):
        # binary serialization in ridiculously small buffer =)
        b_message = list(self.message.batch(10))
        b_raw_message = list(self.raw_message.batch(10))

        # number of batches
        self.assertEqual(len(b_message), 1)
        self.assertEqual(len(b_raw_message), 2)

        # lets read stuff back. number of sent before ID's is of course 0.
        self.check_message(b_message[0], 0, self.message)
        self.check_message(b_raw_message[0], 0, self.raw_message)
        self.check_message(b_raw_message[1], 1, self.raw_message)

    def check_message(self, batch, itr, msg):
        sent, data = batch
        # we send batches of 1 token size
        self.assertEqual(sent, itr)
        # |COMMAND|FRAME-LEN|{token}|{payload}|{id:4}|{expiry:4}|{priority:1}
        command, frame_len = struct.unpack(">BI", data[0:5])
        self.assertEqual(command, 2)
        self.assertEqual(frame_len, len(data) - 5)
        
        off = 5
        restored = {}
        for itm in range(1, 6):
            hdr, length = struct.unpack(">BH", data[off:(off+3)])
            off += 3
            value = data[off:(off+length)]
            off += length
            if hdr == 1:
                restored['token'] = binascii.hexlify(value).decode('ascii')
            elif hdr == 2:
                restored['payload'] = json.loads(value.decode('utf-8'))
            elif hdr == 3:
                restored['index'] = struct.unpack(">I", value)[0]
            elif hdr == 4:
                restored['expiry'] = struct.unpack(">I", value)[0]
            elif hdr == 5:
                restored['priority'] = struct.unpack(">B", value)[0]

        for key in ('token', 'payload', 'index', 'expiry', 'priority'):
            if key not in restored:
                self.fail("Binary message is missing: %s" % key)

        # check message
        self.assertEqual(msg.tokens[itr].lower(), restored['token'].lower())
        self.assertEqual(msg.payload['aps'], restored['payload']['aps'])
        restored['payload'].pop('aps')
        self.assertEqual(msg.extra, restored['payload'])
        self.assertEqual(restored['index'], itr)
        self.assertEqual(msg.expiry, restored['expiry'])
        self.assertEqual(msg.priority, restored['priority'])

    def test_retry(self):
        # include failed
        r_message = self.message.retry(0, True)
        for key in ('tokens', 'alert', 'badge', 'sound', 'content_available', 'expiry', 'priority', 'extra'):
            self.assertEqual(getattr(self.message, key), getattr(r_message, key))

        # nothing to retry, we skip the token
        self.assertEqual(self.message.retry(0, False), None)

        # include failed
        r_raw_message = self.raw_message.retry(0, True)
        for key in ('tokens', 'alert', 'badge', 'sound', 'content_available', 'expiry', 'priority', 'extra'):
            self.assertEqual(getattr(self.raw_message, key), getattr(r_raw_message, key))

        # skip failed
        r_raw_message = self.raw_message.retry(0, False)
        self.assertEqual(self.raw_message.tokens[1:], r_raw_message.tokens)
        for key in ('alert', 'badge', 'sound', 'content_available', 'expiry', 'priority', 'extra'):
            self.assertEqual(getattr(self.raw_message, key), getattr(r_raw_message, key))


class APNsClientResultTest(unittest.TestCase):
    """ Test Result API. """

    def setUp(self):
        self.msg = Message(["0123456789ABCDEF", "FEDCBA9876543210"], alert="message")

    def test_result(self):
        for reason in Result.ERROR_CODES.keys():
            res = Result(self.msg, (reason, 0))
            self.assertEqual(len(res.errors), int(reason in (1, 3, 4, 6, 7, None)))
            self.assertEqual(len(res.failed), int(reason in (2, 5, 8)))
            self.assertEqual(reason in (1, 2, 5, 8, 10, None), res.needs_retry())

            if res.needs_retry():
                ret = res.retry()
                # skip failed or successful token by Shutdown
                self.assertEqual(len(ret.tokens), 2 - len(res.failed) - int(reason == 10))


class APNsDummyTest(unittest.TestCase):
    """ Test APNs client with dummy backend. """

    def test_send(self):
        # success, retry + include-failed, don't-retry + include-failed
        backend = DummyBackend(push=(None, 1, 3))
        session = Session(pool=backend)

        msg = Message(["0123456789ABCDEF", "FEDCBA9876543210"], alert="my alert", badge=10, content_available=1, my_extra=15)
        push_con = session.get_connection("push_production", cert_string="certificate")
        srv = APNs(push_con)
        res = srv.send(msg)
        self.assertEqual(len(res.failed), 0)
        self.assertEqual(len(res.errors), 0)
        self.assertFalse(res.needs_retry())

        push_con2 = session.get_connection("push_production", cert_string="certificate")
        srv = APNs(push_con2)
        self.assertEqual(session.pool.push_result_pos, 0)
        session.pool.push_result_pos += 1
        res = srv.send(msg)
        self.assertEqual(len(res.failed), 0)
        self.assertEqual(len(res.errors), 1)
        self.assertTrue(res.needs_retry())

        # indeed, we have used the cache
        self.assertEqual(session.pool.new_connections, 1)

        push_con = session.new_connection("push_production", cert_string="certificate")
        srv = APNs(push_con)
        res = srv.send(msg)
        self.assertEqual(len(res.failed), 0)
        self.assertEqual(len(res.errors), 1)
        self.assertFalse(res.needs_retry())

        # indeed, new connection, we haven't used the cache
        self.assertEqual(session.pool.new_connections, 2)

    def test_feedback(self):
        backend = DummyBackend(feedback=5)
        session = Session(pool=backend)
        feed_con = session.new_connection("feedback_production", cert_string="certificate")
        srv = APNs(feed_con)
        self.assertEqual(len(list(srv.feedback())), 5)


if __name__ == '__main__':
    unittest.main()



# Copyright 2014 Sardar Yumatov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__title__ = 'APNS client'
__version__ = "0.2"
__author__ = "Sardar Yumatov"
__contact__ = "ja.doma@gmail.com"
__license__ = "Apache 2.0"
__homepage__ = "https://bitbucket.org/sardarnl/apns-client/"
__copyright__ = 'Copyright 2014 Sardar Yumatov'


from apnsclient.apns import APNs, Message
from apnsclient.transport import Session



# Copyright 2014 Sardar Yumatov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import select
import socket
import OpenSSL

try:
    import threading as _threading
except ImportError:
    import dummy_threading as _threading

# python 3 support
import six

from . import BaseBackend, BaseConnection
from ..certificate import BaseCertificate

# module level logger
LOG = logging.getLogger(__name__)


class Certificate(BaseCertificate):
    """ pyOpenSSL certificate implementation. """

    def load_context(self, cert_string=None, cert_file=None, key_string=None, key_file=None, passphrase=None):
        """ Initialize and load certificate context. """
        context = OpenSSL.SSL.Context(OpenSSL.SSL.SSLv3_METHOD)
        if passphrase is not None and not isinstance(passphrase, six.binary_type):
            passphrase = six.b(passphrase)
        
        if cert_file:
            # we have to load certificate for equality check. there is no
            # other way to obtain certificate from context.
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Certificate provided as file: %s", cert_file)

            with open(cert_file, 'rb') as fp:
                cert_string = fp.read()
        else:
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Certificate provided as string")

        cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, cert_string)
        context.use_certificate(cert)

        if not key_string and not key_file:
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Private key provided with certificate %s %s passphrase",
                            'file' if cert_file else 'string',
                            'with' if passphrase is not None else 'without')

            # OpenSSL is smart enought to locate private key in the certificate
            args = [OpenSSL.crypto.FILETYPE_PEM, cert_string]
            if passphrase is not None:
                args.append(passphrase)

            pk = OpenSSL.crypto.load_privatekey(*args)
            context.use_privatekey(pk)
        elif key_file and passphrase is None:
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Private key provided as file without passphrase: %s", key_file)

            context.use_privatekey_file(key_file, OpenSSL.crypto.FILETYPE_PEM)
        else:
            if key_file:
                if LOG.isEnabledFor(logging.DEBUG):
                    LOG.debug("Private key provided as file withpassphrase: %s", key_file)

                # key file is provided with passphrase. context.use_privatekey_file
                # does not use passphrase, so we have to load the key file manually.
                with open(key_file, 'rb') as fp:
                    key_string = fp.read()
            else:
                if LOG.isEnabledFor(logging.DEBUG):
                    LOG.debug("Private key provided as string %s passphrase",
                                'with' if passphrase is not None else 'without')

            args = [OpenSSL.crypto.FILETYPE_PEM, key_string]
            if passphrase is not None:
                args.append(passphrase)

            pk = OpenSSL.crypto.load_privatekey(*args)
            context.use_privatekey(pk)

        # check if we are not passed some garbage
        context.check_privatekey()
        return context, cert

    def dump_certificate(self, raw_certificate):
        """ Dump certificate as PEM string.
        
            :Arguments:
                - context (object): certificate context as returned by :func:`load_context`

            :Returns:
                Certificate string as sequence of bytes.
        """
        return OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, raw_certificate)

    def dump_digest(self, raw_certificate, digest):
        """ Dump certificate digest.

            :Arguments:
                - context (object): certificate context as returned by :func:`load_context`
                - digest (str): digest name, such as "sha1"

            :Returns:
                Digest as sequence of bytes.
        """
        return raw_certificate.digest(digest)


class Connection(BaseConnection):
    """ Connection to APNs. """

    def __init__(self, address, certificate, timeout=None):
        """ Open new connection to APNs using POSIX sockets and pyOpenSSL.

            :Arguments:
                - address (tuple): address as (host, port) tuple.
                - certificate (:class:`Certificate`): provider's certificate.
                - timeout (float): connection timeout in seconds.
        """
        super(Connection, self).__init__(address, certificate)
        self._socket = None
        self._connection = None
        self._timeout = timeout
        self._open_connection(timeout)

    def _open_connection(self, timeout):
        """ Open remote connection. """
        try:
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Opening POSIX socket/pyOpenSSL connection to %r", self.address)
            self._socket = self._create_socket(timeout)
            self._configure_socket()
            self._connection = self._create_openssl_connection()
            self._configure_connection()
            self._connect_and_handshake()
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("SSL handshaking to %r completed", self.address)
        except:
            LOG.warning("Failed to establish socket/SSL connection to %r", self.address, exc_info=True)
            self.close()
            raise

    def _create_socket(self, timeout):
        """ Create new plain TCP socket. """
        return socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def _configure_socket(self):
        """ Hook to configure socket parameters. """
        pass

    def _create_openssl_connection(self):
        """ Create new OpenSSL connection. """
        ctx = self.certificate.get_context()
        if self._timeout is not None:
            ctx.set_timeout(int(self._timeout))
        return OpenSSL.SSL.Connection(ctx, self._socket)

    def _configure_connection(self):
        """ Hookt to configure SSL connection. """
        pass

    def _connect_and_handshake(self):
        """ Connect to APNs and SSL handshake. """
        self._connection.connect(self.address)
        # NOTE: if we set any timeout on the underlying socket, then OpenSSL wrapper
        # will always immediately fail with WantReadError without even attempting
        # to block for that timeout. So we use an infinitelly blocking socket here,
        # but we have set timeout on context object and we hope it works.
        self._connection.setblocking(1)
        self._connection.do_handshake()

    def closed(self):
        """ Returns True if this connection is explicitly closed with :func:`close` call. """
        return self._socket is None

    def close(self):
        """ Close this connection. """
        if self._socket is not None:
            if self._connection is not None:
                try:
                    # tell SSL socket we are done
                    self._connection.shutdown()
                except:
                    pass
                try:
                    # free SSL related resources
                    self._connection.close()
                except:
                    pass

            try:
                # shutdown IO
                self._socket.shutdown(socket.SHUT_RDWR)
            except:
                pass
            try:
                # close socket
                self._socket.close()
            except:
                pass

            self._socket = None
            self._connection = None

    def reset(self):
        """ Flushes read buffer. """
        to_skip = self._connection.pending()
        if to_skip > 0:
            self._connection.recv(to_skip)

    def write(self, data, timeout):
        """ Write chunk of data. """
        if self.closed():
            raise IOError("Connection is closed")

        # FIXME: pyOpenSSL can't block with the socket if timeout is set to
        # any value except None. So it is either block forewer or immediate
        # WantReadError without waiting at all. we have to use select() for
        # blocking.
        self._connection.setblocking(1)
        self._socket.settimeout(timeout)  # None will set blocking mode
        waited = 0
        while True:
            try:
                # usually socket accepts any data regardless of network state.
                # it simply stays in the output buffer, which we can't flush
                # unless we disable Nagle algorithm entirely.
                if timeout is not None:
                    before = time.time()
                    sent = self._connection.send(data)
                    waited += time.time() - before
                else: # save the time syscalls
                    sent = self._connection.send(data)

                if sent == len(data):
                    return
                elif sent > 0:
                    data = data[sent:]
                    if timeout is not None:
                        # in case pyOpenSSL starts to actually wait timeouts
                        if timeout - waited >= 0:
                            self._socket.settimeout(timeout - waited)
                            # updated timeout, try again
                            continue
                    else:
                        # infinite timeout, keep trying
                        continue
            except OpenSSL.SSL.WantWriteError:
                # looks like we are in blocking mode, but pyOpenSSL doesn't want
                # to block or we are in a really short timeout.
                if timeout is not None:
                    waited += time.time() - before
                    if timeout - waited > 0:
                        before = time.time()
                        _, canwrite, _ = select.select((), (self._socket, ), (), timeout - waited)
                        waited += time.time() - before
                        if canwrite and timeout - waited >= 0:
                            # buffer ready for writing, try again
                            self._socket.settimeout(timeout - waited)
                            continue

            # we come here if send() blocking works and we waited more than timeout
            # or send() blocking doesn't work and we waited more than timeout
            # or if timeout is None and we still got WantWriteError
            LOG.warning("Write timeout %r it too short for buffer %s", timeout, len(data))
            raise IOError("Timeout exceeded")

    def peek(self, size):
        """ Peek chunk of data from the read buffer. """
        if self.closed():
            return None

        self._connection.setblocking(1)
        pending = self._connection.pending()
        if pending > 0:
            return self._connection.recv(min(pending, size))
        else:
            # do not trust pyOpenSSL, that piece of crap never sets pending buffer.
            canread, _, _ = select.select((self._socket, ), (), (), 0) # poll
            if canread:
                self._connection.setblocking(0)
                self._socket.settimeout(0)
                try:
                    ret = self._connection.recv(size)
                    if not ret:
                        ret = None

                    return ret
                except OpenSSL.SSL.ZeroReturnError:
                    # nice end of stream
                    return None
                except OpenSSL.SSL.WantReadError:
                    # no data is available yet
                    return None

        return None

    def read(self, size, timeout):
        """ Read chunk of data. """
        if self.closed():
            return None

        self._connection.setblocking(1)
        self._socket.settimeout(timeout)  # None will set blocking mode
        waited = 0
        while True:
            try:
                if timeout is not None:
                    before = time.time()
                    ret = self._connection.recv(size)
                    waited += time.time() - before
                else: # save the time syscalls
                    ret = self._connection.recv(size)

                if not ret:
                    # in case recv() responds with empty string
                    ret = None
                
                return ret
            except OpenSSL.SSL.ZeroReturnError:
                # nice end of stream
                return None
            except OpenSSL.SSL.WantReadError:
                # either pyOpenSSL is failing without even trying to block
                # or we really exceeded the timeout
                if timeout is not None:
                    waited += time.time() - before
                    if timeout - waited > 0:
                        before = time.time()
                        canread, _, _ = select.select((self._socket, ), (), (), timeout - waited)
                        waited += time.time() - before
                        if canread and timeout - waited >= 0:
                            # buffer ready for writing, try again
                            self._socket.settimeout(timeout - waited)
                            continue

            # we come here if read raises WantReadError, which means normal blocking
            # doesn't work or we really exceeded the timeout; timeout is None
            # and we still got WantReadError, which is strange; we waited all
            # time with select()
            #LOG.warning("Read timeout %r it too short for buffer %s", timeout, size)
            raise IOError("Timeout exceeded")


class Backend(BaseBackend):
    """ IO backend based on POSIX sockets. """
    # no, we can't detect if socket is closed without IO
    can_detect_close = False
    # raw connection implementation
    connection_class = Connection

    def get_new_connection(self, address, certificate, timeout=None):
        """ Open a new connection.
        
            :Arguments:
                - address (tuple): target (host, port).
                - certificate (:class:`Certificate`): certificate instance.
                - timeout (float): connection timeout in seconds
        """
        return self.connection_class(address, certificate, timeout=timeout)

    def get_certificate(self, cert_params):
        """ Create/load certificate from parameters. """
        return Certificate(**cert_params)

    def create_lock(self):
        """ Provides semaphore with ``threading.Lock`` interface. """
        return _threading.Lock()



# Copyright 2014 Sardar Yumatov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from struct import pack

try:
    import threading as _threading
except ImportError:
    import dummy_threading as _threading

from . import BaseBackend, BaseConnection
from ..certificate import BaseCertificate

# python 3 support
import six

# module level logger
LOG = logging.getLogger(__name__)


class Certificate(BaseCertificate):
    """ Dummy certificate """

    def load_context(self, cert_string=None, cert_file=None, key_string=None, key_file=None, passphrase=None):
        """ Returns None as we don't handle any context. """
        return None, None

    def dump_certificate(self, raw_certificate):
        """ Returns dummy contents. All dummy certificates are equal. """
        return "CERTIFICATE"

    def dump_digest(self, raw_certificate, digest):
        """ Returns dummy digsst. All dummy certificates are equal. """
        return self.dump_certificate(raw_certificate)


class Backend(BaseBackend):
    """ Dummy backend designed for testing without performing real IO. Serves
        as an exmple for your custom backends.
    """
    # simulate stdio behavior
    can_detect_close = False

    def __init__(self, push=None, feedback=None, **options):
        """ Create new backend.
        
            :Arguments:
                - push (list): list of status codes to return while sending messages.
                - feedback (int): number of tokens to generate in the feedback stream.
        """
        super(Backend, self).__init__(**options)
        self.push_results = push
        self.push_result_pos = -1
        self.feedback_results = feedback
        self.new_connections = 0
        assert (push is not None) ^ (feedback is not None), "Push results or feedback stream must be provided"

    def get_new_connection(self, address, certificate, timeout=None):
        """ Open a new connection.
        
            :Arguments:
                - address (tuple): target (host, port).
                - certificate (:class:`Certificate`): certificate instance.
                - timeout (float): connection timeout in seconds
        """
        self.new_connections += 1
        self.push_result_pos += 1
        return Connection(self, address, certificate)

    def get_certificate(self, cert_params):
        """ Create/load certificate from parameters. """
        return Certificate(**cert_params)

    def create_lock(self):
        """ Provides semaphore with ``threading.Lock`` interface. """
        return _threading.Lock()


class Connection(BaseConnection):
    """ Dummy connection. """

    def __init__(self, pool, address, certificate):
        """ Create new dummy connection.
        
            :Arguments:
                - pool (:class:`Backend`): dummy backend.
                - address (tuple): target host and port.
                - certificate (:class:`Certificate`): provider certificate.
        """
        super(Connection, self).__init__(address, certificate)
        self.pool = pool
        self._closed = False

    def closed(self):
        """ Returns True if :func:`close` has been explicitly called. """
        return self._closed

    def close(self):
        """ Marks this connection as closed. """
        self._closed = True

    def reset(self):
        """ Reset dummy connection to use next result record. """
        pass

    def write(self, data, timeout):
        """ Does nothing, always succeeds. """
        if self.closed():
            raise IOError("Connection closed")
    
    def peek(self, size):
        """ Always returns None as we never fail prematurely. """
        return None

    def read(self, size, timeout):
        """ Iterates over preconfigured send/feedback responses. """
        if self.closed():
            return None

        if self.pool.push_results is not None:
            # we are push connection
            ret = self.pool.push_results[self.pool.push_result_pos % len(self.pool.push_results)]
            if ret is not None:
                ret = pack(">BBI", 8, ret, 0)

            return ret
        else: # feedback mode
            ret = []
            for x in range(0, self.pool.feedback_results):
                token = six.b("test_{}".format(x))
                ret.append(pack(">IH{}s".format(len(token)), int(time.time()), len(token), token))

            self.close()
            return six.binary_type().join(ret)



# Copyright 2014 Sardar Yumatov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ('BaseCertificate', )


class BaseCertificate(object):
    """ Default certificate loader. """
    # The way how to compare certificates. If None, then certificate body will
    # be dumped as bytestring and compared. Other option is to provide digest
    # method supported by your SSL implementation. Certificates are considered
    # equal if their digests are equal.
    equality_check = 'sha1'

    def __init__(self, cert_string=None, cert_file=None, key_string=None, key_file=None, passphrase=None):
        """ Provider's certificate and private key.
        
            Your certificate will probably contain the private key. Open it
            with any text editor, it should be a plain text (PEM format). The
            certificate is enclosed in ``BEGIN/END CERTIFICATE`` strings and
            private key is in ``BEGIN/END RSA PRIVATE KEY`` section. If you can
            not find the private key in your .pem file, then you should
            provide it with `key_string` or `key_file` argument.

            :Arguments:
                - cert_string (bytes): certificate in PEM format from string.
                - cert_file (str): certificate in PEM format from file.
                - key_string (bytes): private key in PEM format from string.
                - key_file (str): private key in PEM format from file.
                - passphrase (bytes): passphrase for your private key.
        """
        self.context, cert = self.load_context(
            cert_string=cert_string,
            cert_file=cert_file,
            key_string=key_string,
            passphrase=passphrase
        )
        if self.equality_check is None:
            self._equality = self.dump_certificate(cert)
        else:
            self._equality = self.dump_digest(cert, self.equality_check)

    def load_context(self, cert_string=None, cert_file=None, key_string=None, key_file=None, passphrase=None):
        """ Initialize and load certificate context.
            
            :Returns:
                (context, raw_certificate) the SSL context and raw certificate instance.
        """
        raise NotImplementedError

    def dump_certificate(self, raw_certificate):
        """ Dump certificate as PEM string.
        
            :Arguments:
                - raw_certificate (object): raw certificate as returned by :func:`load_context`

            :Returns:
                Certificate string as sequence of bytes.
        """
        raise NotImplementedError

    def dump_digest(self, raw_certificate, digest):
        """ Dump certificate digest.

            :Arguments:
                - raw_certificate (object): raw certificate as returned by :func:`load_context`
                - digest (str): digest name, such as "sha1"

            :Returns:
                Digest as sequence of bytes.
        """
        raise NotImplementedError

    def get_context(self):
        """ Returns SSL context instance. You may use that context to specify
            required verification level, trusted CA's etc.
        """
        return self.context

    def __hash__(self):
        """ Content based hash. """
        return hash(self._equality)

    def __eq__(self, other):
        """ True if other object is Certificate instance and it passes the equality check. """
        if hasattr(other, "_equality") and hasattr(other, "equality_check"):
            return self.equality_check == other.equality_check and self._equality == other._equality

        return False



# Copyright 2014 Sardar Yumatov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import binascii
import datetime
from struct import unpack

# python 3 support
import six

__all__ = ('Session', 'Connection')

# module level logger
LOG = logging.getLogger(__name__)


class Session(object):
    """ The front-end for the underlying connection pool. """
    # Default APNs addresses.
    ADDRESSES = {
        "push_sandbox": ("gateway.sandbox.push.apple.com", 2195),
        "push_production": ("gateway.push.apple.com", 2195),
        "feedback_sandbox": ("feedback.sandbox.push.apple.com", 2196),
        "feedback_production": ("feedback.push.apple.com", 2196),
    }
    # Default timeout for attempting a new connection.
    DEFAULT_CONNECT_TIMEOUT = 10
    # Default write buffer size. Should be close to MTU size.
    DEFAULT_WRITE_BUFFER_SIZE = 2048
    # Default timeout for write operations.
    DEFAULT_WRITE_TIMEOUT = 20
    # Default read buffer size, used by feedback.
    DEFAULT_READ_BUFFER_SIZE = 2048
    # Default timeout for read operations.
    DEFAULT_READ_TIMEOUT = 20
    # Default timeout waiting for error response at the end message send operation.
    DEFAULT_READ_TAIL_TIMEOUT = 3
    
    def __init__(self, pool="apnsclient.backends.stdio",
                       connect_timeout=DEFAULT_CONNECT_TIMEOUT,
                       write_buffer_size=DEFAULT_WRITE_BUFFER_SIZE,
                       write_timeout=DEFAULT_WRITE_TIMEOUT,
                       read_buffer_size=DEFAULT_READ_BUFFER_SIZE,
                       read_timeout=DEFAULT_READ_TIMEOUT,
                       read_tail_timeout=DEFAULT_READ_TAIL_TIMEOUT,
                       **pool_options):
        """ The front-end to the underlying connection pool. The purpose of this
            class is to hide the transport implementation that is being used for
            networking. Default implementation uses built-in python sockets and
            ``select`` for asynchronous IO.

            :Arguments:
                - pool (str, type or object): networking layer implementation.
                - connect_timeout (float): timeout for new connections.
                - write_buffer_size (int): chunk size for sending the message.
                - write_timeout (float): maximum time to send single chunk in seconds.
                - read_buffer_size (int): feedback buffer size for reading.
                - read_timeout (float): timeout for reading single feedback block.
                - read_tail_timeout (float): timeout for reading status frame after message is sent.
                - pool_options (kwargs): passed as-is to the pool class on instantiation.
        """
        # IO deafults
        self.connect_timeout = connect_timeout
        self.write_buffer_size = write_buffer_size
        self.write_timeout = write_timeout
        self.read_buffer_size = read_buffer_size
        self.read_timeout = read_timeout
        self.read_tail_timeout = read_tail_timeout

        # class name given by qualified name
        if isinstance(pool, six.string_types):
            pool_module = __import__(pool)
            for name in pool.split('.')[1:]:
                try:
                    pool_module = getattr(pool_module, name)
                except AttributeError:
                    raise ImportError("Can't load pool backend", pool)

            try:
                pool = getattr(pool_module, "Backend")
            except AttributeError:
                raise ImportError("Can't find Backend class in pool module", pool)

        # resolved or given as class
        if isinstance(pool, type):
            pool = pool(**pool_options)

        self.pool = pool
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug("New session, WB: %sb/%ss, RB: %sb/%ss, TT: %ss, Pool: %s",
                      write_buffer_size, write_timeout,
                      read_buffer_size, read_timeout,
                      read_tail_timeout,
                      pool.__class__.__module__)

    @classmethod
    def get_address(cls, address):
        """ Maps address to (host, port) tuple. """
        if not isinstance(address, (list, tuple)):
            addr = cls.ADDRESSES.get(address)
            if addr is None:
                raise ValueError("Unknown address mapping: {0}".format(address))

            address = addr
        return address

    def new_connection(self, address="feedback_sandbox", certificate=None, **cert_params):
        """ Obtain new connection to APNs. This method will not re-use existing
            connection from the pool. The connection will be closed after use.
            
            Unlike :func:`get_connection` this method does not cache the
            connection.  Use it to fetch feedback from APNs and then close when
            you are done.

            :Arguments:
                - address (str or tuple): target address.
                - certificate (:class:`BaseCertificate`): provider's certificate instance.
                - cert_params (kwargs): :class:`BaseCertificate` arguments, used if ``certificate`` instance is not given.
        """
        if certificate is not None:
            cert = certificate
        else:
            cert = self.pool.get_certificate(cert_params)

        address = self.get_address(address)
        return Connection(address, cert, self, use_cache=False)

    def get_connection(self, address="push_sanbox", certificate=None, **cert_params):
        """ Obtain cached connection to APNs.

            Session caches connection descriptors, that remain open after use.
            Caching saves SSL handshaking time. Handshaking is lazy, it will be
            performed on first message send.

            You can provide APNs address as ``(hostname, port)`` tuple or as
            one of the strings:

                - push_sanbox -- ``("gateway.sandbox.push.apple.com", 2195)``, the default.
                - push_production -- ``("gateway.push.apple.com", 2195)``
                - feedback_sandbox -- ``("feedback.sandbox.push.apple.com", 2196)``
                - feedback_production -- ``("feedback.push.apple.com", 2196)``

            :Arguments:
                - address (str or tuple): target address.
                - certificate (:class:`BaseCertificate`): provider's certificate instance.
                - cert_params (kwargs): :class:`BaseCertificate` arguments, used if ``certificate`` instance is not given.
        """
        if certificate is not None:
            cert = certificate
        else:
            cert = self.pool.get_certificate(cert_params)

        address = self.get_address(address)
        return Connection(address, cert, self, use_cache=True)

    def outdate(self, delta):
        """ Close open unused connections in the pool that are left untouched
            for more than ``delta`` time.

            You may call this method in a separate thread or run it in some
            periodic task. If you don't, then all connections will remain open
            until session is shut down. It might be an issue if you care about
            your open server connections.

            :Arguments:
                delta (``timedelta``): maximum age of unused connection.
        """
        if LOG.isEnabledFor(logging.DEBUG):
            if delta.total_seconds() == 0.0:
                LOG.debug("Shutdown session")
            else:
                LOG.debug("Outdating session with delta: %s", delta)

        self.pool.outdate(delta)

    def shutdown(self):
        """ Shutdown all connections in the pool. This method does will not close
            connections being use at the calling time.
        """
        # NOTE: global package datetime can become None if session is stored in
        # a global variable and being garbage collected with the rest of the module.
        if datetime is not None:
            self.pool.outdate(datetime.timedelta())

    def __del__(self):
        """ Last chance to shutdown() """
        self.shutdown()


class Connection(object):
    """ Connection wrapper. """

    def __init__(self, address, certificate, session, use_cache=False):
        """ New connection wrapper.
            
            :Arguments:
                - address (tuple) - (host, port) to connect to.
                - certificate (:class:`BaseCertificate`) - provider certificate.
                - session (object) - parent session.
                - use_cache (bool) - True if connections may be cached in the pool.
        """
        self.address = address
        self.certificate = certificate
        self.session = session
        self.use_cache = use_cache
        self._reused = False
        self._connection = None
        self._lock = self.session.pool.create_lock()

    def __enter__(self):
        try:
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Entering networking session")

            self._lock.acquire() # block until lock is given
            self._open_connection() # can raise exception, bubblit up to the top
        except:
            self._lock.release()
            raise

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Leaving networking session%s%s",
                          " with still open connection" if self._connection else "",
                          " because of failure" if exc_type else "",
                          exc_info=(exc_type is not None))

            # the only possible scenario when connection is left open while
            # we are here is when some totally unexpected exception bubbles up.
            assert exc_type is not None or self._connection is None
            if self._connection:
                # terminate connection in case of an error
                self._close(terminate=True)
        finally:
            self._lock.release()

        # we return None, which is False, which forces python to re-raise on error

    def send(self, message):
        """ Send message. """
        # will raise exception if non-cached connection has been used
        with self:
            batch = message.batch(self.session.write_buffer_size)
            failed_after = None
            status = None
            total_sent = 0
            decoder = ResponseDecoder()
            for iteration, (sent, chunk) in enumerate(batch):
                assert len(chunk) > 0
                total_sent += len(chunk)

                if iteration == 0:
                    # can raise exception if new connection is failed to open.
                    # write doesn't return anything, but we are interested in IO failures.
                    _, io_exception = self._ensuring_io(lambda con: con.write(chunk, self.session.write_timeout))
                    if io_exception is not None:
                        # IO failure on first write, sent is 0 here, retry with
                        # the whole message
                        failed_after = sent
                        break
                else:
                    # other writes, that fail naturally
                    try:
                        ret = self._connection.write(chunk, self.session.write_timeout)
                    except:
                        # IO failure on subsequent writes, some of the tokens are
                        # sent, break on the beginning of this batch
                        failed_after = sent
                        break

                # check for possibly arriving failure frame
                try:
                    # should either return sequence of bytes or None if read buffer
                    # is empty.
                    ret = self._connection.peek(256) # status frame is 6 bytes
                except:
                    # Peek failed, which means our read operations fail
                    # abnormaly. I don't like that and the final read will
                    # probably fail too. So fail early, possibly messing the
                    # first batch, but not everything
                    failed_after = sent
                    break
                else:
                    if ret is not None:
                        decoder.feed(ret)
                        # status is not None only if previous iteration got
                        # successful status, but it appears not to be the last
                        # chunk. this should not happen by APNs protocol, still
                        # we have to handle it. the easy solution: ignore
                        # previous status (probably garbage in read buffer)
                        # with a warning.
                        if status is not None:
                            LOG.warning("Got success frame while batch is not comleted. Frame ignored.")

                        # NOTE: it is possible we get None here because not all
                        # bytes could be read without blocking. on next iteration
                        # or final blocking read we will get the rest of the bytes.
                        status = decoder.decode()
                        if status is not None and status[0] != 0: # error detected
                            if LOG.isEnabledFor(logging.INFO):
                                LOG.info("Message send failed midway with status %r to address %r. Sent tokens: %s, bytes: %s",
                                         status, self.address, sent, total_sent)

                            # some shit had happened, response from APNs, bail out and prepare for retry
                            self._close(terminate=True)
                            return status
                    # else: nothing in the read buffer, keep sending

            # by this time we either stopped prematurely on IO error with
            # failed_after set or we finished all batches, possibly having
            # status read with non-blocking IO.

            # the write stream possibly failed, but the read stream might be still
            # open with status frame precisely pointing to failed token.
            if status is None:
                # read status frame, could take 2 iterations if the fist one returns
                # just the read buffer with few bytes not making the whole status frame.
                while True:
                    try:
                        ret = self._connection.read(256, self.session.read_tail_timeout)
                    except:
                        # one of two things had happened:
                        #  - everything went fine, we waited for the final status
                        #    frame the tail timeout of time and got nothing (timeout).
                        #    this is a success condition according to APNs documentation
                        #    if status frame with code 0 is not sent (it is never sent).
                        #  - reading failed with some other exception. we don't know
                        #    starting from which token the batch has failed. we can't
                        #    attempt to read status frame again, because read stream
                        #    is probably closed by now. there is nothing we can do
                        #    except pretending everything is OK. the failed tokens
                        #    will be reported by feedback. the tokens that didn't got
                        #    the message... well, so is life, we can't detect them.
                        #
                        # Sorry, but this is how APNs protocol designed, I can't
                        # do better here. APNs developers suck hard.
                        #
                        # We still have to check failed_after, it tells us when
                        # IO write failed. If failed_after is not None, then
                        # we got here probably because connection is closed for
                        # read and write after the write failure.
                        break
                    else:
                        if ret is not None:
                            decoder.feed(ret)
                            status = decoder.decode()
                            if status is None:
                                # we got bytes, but not enogugh for the status frame.
                                continue

                            # complete status frame read, evaluate
                            if status[0] != 0:
                                if LOG.isEnabledFor(logging.INFO):
                                    LOG.info("Message send failed with status %r to address %r. Sent tokens: %s, bytes: %s",
                                             status, self.address, len(message.tokens), total_sent)

                                # some shit had indeed happened
                                self._close(terminate=True)
                                return status

                        # got a successful status or read ended with closed connection
                        break

            # by this time we have either successful status frame (code 0) or
            # we failed to obtain status frame at all. the failed_after is not None
            # if IO write failed before.

            # there are some bytes read, but we failed to read complete status
            # frame.  all possible timeouts are exceeded or read stream is
            # totally fucked up, so we can't wait and read again. let the user
            # know this happened and treat the situation as if no frame was
            # received at all. APNs protocol sucks sooo much.
            if status is None and decoder._buf:
                LOG.warning("Failed to read complete status frame from %r, but has read some bytes before. Probably read timeout %s is to short.",
                            self.address, self.session.read_tail_timeout)

                # close connection, it is failing
                self._close(terminate=True)

            # normall successs scenario with success frame provided. never
            # happens according to APNs documentation (no status frame gets
            # sent on success), but can happen logically.
            if failed_after is None and status is not None and status[0] == 0:
                # success, release connection for re-use if it was meant for reuse
                self._close(terminate=not self._reused)
                return status

            # everything looks like success, but it might be because read stream
            # was closed or just timeouted. check write IO failure.
            if failed_after is not None:
                if LOG.isEnabledFor(logging.INFO):
                    LOG.info("Message send failed midway with status %r to address %r. Sent tokens: %s, bytes: %s",
                             status, self.address, failed_after, total_sent)

                # close connection, it is failing
                self._close(terminate=True)
                return (255, failed_after + 1)

            # we have sent message to all target tokens and have waited for
            # tail_timeout for any error reponse to arrive. Nothing arrived
            # (hopefully not because read error) and we did not fail with write
            # failure middle on the road, so according to Apple's manual
            # everything went OK. This protocol sucks.
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Message sent successfully to address %r. Sent tokens: %s, bytes: %s",
                         self.address, len(message.tokens), total_sent)

            # success, release connection for re-use if it was meant for reuse
            self._close(terminate=not self._reused)
            return None

    def feedback(self):
        """ Read and parse feedback information. """
        if self.use_cache:
            # sanity check
            LOG.warning("Don't use cached connections for feedback, you might get stale data.")

        # will raise exception if non-cached connection has been used
        with self:
            # on connection failure we bubble up the exceptions. on IO failure
            # we get the exception as return value, stopping the iteration normally.
            data, io_exception = self._ensuring_io(lambda con: con.read(self.session.read_buffer_size, self.session.read_timeout))
            # data is non empty sequence of bytes on success, None if connection
            # has been closed or on failure. io_exception is not None on IO errors.

            feedback = FeedbackDecoder()
            total_records = 0
            failed = io_exception is not None
            while data is not None:
                feedback.feed(data)
                # TODO: use yield from
                for record in feedback.decoded():
                    total_records += 1
                    yield record

                try:
                    # read next chunk, leaving again either sequence of bytes or
                    # None if connection has been closed.
                    data = self._connection.read(self.session.read_buffer_size, self.session.read_timeout)
                except:
                    # IO failure, probably because of a timeout. break the loop,
                    # we will fetch the rest during the next session.
                    failed = True
                    break

            # there is no point to keep this connection open
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Feedback received %s records from address %r. Stopped %s",
                         total_records, self.address, "by failure" if failed else "successfully")

            # always close feedback connection, preventing stale data
            self._close(terminate=True)

    def _ensuring_io(self, func):
        """ Re-opens connection if read or write has failed. Used to re-initialize
            connections from the pool with a transport not supporting reliable
            socket closed condition.
        """
        failed = False
        if self._reused:
            # if connection is reused, then there might be left over bytes in the
            # read buffer. flush them.
            try:
                self._connection.reset()
            except:
                LOG.info("Failed to reset connection to %r", self.address, exc_info=True)
                # close the connection, prepare for re-connect
                self._close(terminate=True)
                failed = True

        if not failed:
            # OK, reset succeeded or this is a fresh new connetion
            try:
                return func(self._connection), None
            except Exception as exc:
                if self.session.pool.can_detect_close or not self._reused:
                    # bubble up IO related problem on non-cached connection
                    return None, exc

        # Either failed by reset or failed by IO operation. If
        # pool.can_detect_close is False and we are reusing the connection from
        # the cache pool, then it was probably already failing when we got it.
        # Re-get the connection from the pool again.
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug("Re-opening connection to address %r", self.address)

        # ensure failing connection is closed
        self._close(terminate=True)
        # open new connection. this operation might raise exceptions, which
        # will propagate to the outer most caller indicating severe network
        # errors.
        self._open_connection(by_failure=True)
        return self._ensuring_io(func)

    def _open_connection(self, by_failure=False):
        """ Request new connection handle from underlying pool. """
        # use pool if caching is requested or we are ensuring connection with
        # cache enabled.
        if self.use_cache and (not by_failure or self.session.pool.use_cache_for_reconnects):
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Open cached connection to %r%s.", self.address, " by failure" if by_failure else "")

            self._connection = self.session.pool.get_cached_connection(
                self.address,
                self.certificate,
                timeout=self.session.connect_timeout
            )
            self._reused = True
        else:
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Open new connection to %r%s.", self.address, " by failure" if by_failure else "")

            self._connection = self.session.pool.get_new_connection(
                self.address,
                self.certificate,
                timeout=self.session.connect_timeout
            )
            self._reused = False

    def _close(self, terminate=False):
        """ Close connection. """
        if self._connection:
            if terminate:
                self._connection.close()
            else:
                self.session.pool.release(self._connection)

            self._connection = None


# private
class ResponseDecoder(object):
    """ Response frame decoder. """
    # Response command byte
    COMMAND = 8

    def __init__(self):
        self._buf = []

    def feed(self, data):
        """ Feed next frame with data. """
        self._buf.append(data)

    def decode(self):
        """ Returns reconstructed response frame. """
        buf = six.binary_type().join(self._buf)
        if len(buf) >= 6:
            ret = unpack(">BBI", buf[0:6])
            self._buf = []
            if len(buf) > 6:
                # should normally not happen as there is always a single frame
                self._buf.append(buf[6:])

            assert ret[0] == self.COMMAND, "Got unknown command from APNs: {}. Looks like protocol has been changed.".format(ret[0])
            return (ret[1], ret[2])
        else:
            self._buf = [buf]

        return None


# private
class FeedbackDecoder(object):
    """ Feedback decoder. """

    def __init__(self):
        self._buf = []

    def feed(self, data):
        """ Feed next frame with raw data. """
        self._buf.append(data)

    def decoded(self):
        """ Returns generator over next set of decoded records. """
        buf = six.binary_type().join(self._buf)
        pos = 0
        while (pos + 6) < len(buf):
            timestamp, length = unpack(">IH", buf[pos:(pos + 6)])
            assert length > 0

            if (pos + 6 + length) <= len(buf):
                token = binascii.hexlify(buf[(pos + 6):(pos + 6 + length)])
                pos += 6 + length
                yield token, timestamp
                if pos == len(buf):
                    break
            else:
                break

        # consume everything except suffix
        self._buf=[buf[pos:]]



# Copyright 2014 Sardar Yumatov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

if __name__ == '__main__':
    import os.path, sys
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))

import unittest, datetime
from apnsclient import Session


CERTIFICATE = b"""-----BEGIN CERTIFICATE-----
MIIDPTCCAuegAwIBAgIJALpcHuGZGDLzMA0GCSqGSIb3DQEBBQUAMIGdMQswCQYD
VQQGEwJOTDESMBAGA1UECBMJR3JvbmluZ2VuMRIwEAYDVQQHEwlHcm9uaW5nZW4x
FDASBgNVBAoTC2FwbnMtY2xpZW50MRUwEwYDVQQLEwxjZXJ0aWZpY2F0ZXMxFzAV
BgNVBAMTDlNhcmRhciBZdW1hdG92MSAwHgYJKoZIhvcNAQkBFhFqYS5kb21hQGdt
YWlsLmNvbTAeFw0xNDA1MjUxMjEwMjNaFw0xNDA2MjQxMjEwMjNaMIGdMQswCQYD
VQQGEwJOTDESMBAGA1UECBMJR3JvbmluZ2VuMRIwEAYDVQQHEwlHcm9uaW5nZW4x
FDASBgNVBAoTC2FwbnMtY2xpZW50MRUwEwYDVQQLEwxjZXJ0aWZpY2F0ZXMxFzAV
BgNVBAMTDlNhcmRhciBZdW1hdG92MSAwHgYJKoZIhvcNAQkBFhFqYS5kb21hQGdt
YWlsLmNvbTBcMA0GCSqGSIb3DQEBAQUAA0sAMEgCQQC8hz/oFmBrSXpVwipPxFxZ
qzSGgojSVxM/Mzaf24l/b9oeszfnrq0owk40PG7InSv6l5Z71SXKIOLKwJBgI8Od
AgMBAAGjggEGMIIBAjAdBgNVHQ4EFgQUW0eeuZwjTcf1irE+11/C8fF2tSowgdIG
A1UdIwSByjCBx4AUW0eeuZwjTcf1irE+11/C8fF2tSqhgaOkgaAwgZ0xCzAJBgNV
BAYTAk5MMRIwEAYDVQQIEwlHcm9uaW5nZW4xEjAQBgNVBAcTCUdyb25pbmdlbjEU
MBIGA1UEChMLYXBucy1jbGllbnQxFTATBgNVBAsTDGNlcnRpZmljYXRlczEXMBUG
A1UEAxMOU2FyZGFyIFl1bWF0b3YxIDAeBgkqhkiG9w0BCQEWEWphLmRvbWFAZ21h
aWwuY29tggkAulwe4ZkYMvMwDAYDVR0TBAUwAwEB/zANBgkqhkiG9w0BAQUFAANB
AGwgOsUkTIo1y9v3Y77r02RjJLQRL1P68J8Exiunc53LR97Cg+o7LmaHonaLTUUH
jahmYuZGN/Mty0brOn9hTI0=
-----END CERTIFICATE-----"""

PRIVATE_KEY = b"""-----BEGIN RSA PRIVATE KEY-----
Proc-Type: 4,ENCRYPTED
DEK-Info: DES-EDE3-CBC,3195B956C3AD71F6

g7mfzdit5/26briSho71co+eqHV4Xxw7Y/lJzjuN7SGFjUbh/mBNjOxNmXzVSBWo
50SH/NgyqW7wy4glVLIdZ7u7iR8JZjXeBmH1Aee7+4TLGT5fTDR5aYWW7EPIRJ+E
fAgEm1A0fsiHf7ki++UAFtZTPVcQg1NGeYjkJmpgn4CQwyztcfq1psIvNBBgAdrG
UkHeJvwMalS/cmnVVdH2jR2liCsJhfH4lqH8eUSfGwCBeEjAY0Lg6HPwZY+eiVay
/F5D8oL8GvqrM4hyzxU0KOZ8Pu0VUe1wDXDiNv3TUypZm5Y5sRi9DZ6jwZa7/hiE
h3nbwhSfx2WC3bN8Gs8G9pJV9LGiBY2y2zANYTrq7Mv3iVLkHY8BQIUYQuiUnT20
8jeLSXRfZauDrVW5a8wGSJJLAyHGbXMAMV81z129xf8=
-----END RSA PRIVATE KEY-----"""

PRIVATE_PASS= "test"



class StdIOBackendTest(unittest.TestCase):
    """ Test stdio features. """

    def setUp(self):
        self.session = Session(pool="apnsclient.backends.stdio")

    def test_locking(self):
        """ Test thread locking mechanism """
        lock = self.session.pool.create_lock()
        self.assertIsNotNone(lock)
        self.assertTrue(hasattr(lock, "acquire"))
        self.assertTrue(hasattr(lock, "release"))
        lock.acquire()
        lock.release()

    def test_certificates(self):
        """ Test pyOpenSSL certificates. """
        cert = self.session.pool.get_certificate({
            "cert_string": CERTIFICATE,
            "key_string": PRIVATE_KEY,
            "passphrase": PRIVATE_PASS
        })

        cert2 = self.session.pool.get_certificate({
            "cert_string": (CERTIFICATE + b"\n" + PRIVATE_KEY),
            "passphrase": PRIVATE_PASS
        })

        self.assertIsNotNone(cert.get_context())
        self.assertEqual(cert, cert2)
        
    def test_outdate(self):
        # we are not allowed to do any IO in tests, so no real connections.
        # however, it is good idea to test utility functions even with empty pool.
        self.session.outdate(datetime.timedelta(seconds=60))
        self.session.shutdown()


if __name__ == '__main__':
    unittest.main()



# Copyright 2014 Sardar Yumatov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time
import datetime
import logging
from struct import pack

# python 3 support
import six
import binascii


__all__ = ('APNs', 'Message', 'Result')

# module level logger, defaults to "apnsclient.apns"
LOG = logging.getLogger(__name__)


class APNs(object):
    """ APNs multicaster. """

    def __init__(self, connection):
        """ APNs client.

            :Arguments:
                - connection (:class:`Connection`): the connection to talk to.
        """
        self._connection = connection

    def send(self, message):
        """ Send the message.
        
            The method will block until the whole message is sent. The method
            returns :class:`Result` object, which you can examine for possible
            errors and retry attempts.

            .. note::
                If the client fails to connect to APNs, probably because your
                network is down, then this method will raise the related
                exception. However, if connection is successfully established,
                but later on the IO fails, then this method will prepare a
                retry message with the rest of the failed tokens.

            Example::

                # if you use cached connections, then store this session instance
                # somewhere global, such that it will not be garbage collected
                # after message is sent.
                session = Session()
                # get a cached connection, avoiding unnecessary SSL handshake
                con = session.get_connection("push_production", cert_string=db_certificate)
                message = Message(["token 1", "token 2"], alert="Message")
                service = APNs(con)
                try:
                    result = service.send(message)
                except:
                    print "Check your network, I could not connect to APNs"
                else:
                    for token, (reason, explanation) in result.failed.items():
                        delete_token(token) # stop using that token

                    for reason, explanation in result.errors:
                        pass # handle generic errors

                    if result.needs_retry():
                        # extract failed tokens as new message
                        message = message.retry()
                        # re-schedule task with the new message after some delay

            :Returns:
                :class:`Result` object with operation results.
        """
        if len(message.tokens) == 0:
            LOG.warning("Message without device tokens is ignored")
            return Result(message)

        status = self._connection.send(message)
        return Result(message, status)

    def feedback(self):
        """ Fetch feedback from APNs.

            The method returns generator of ``(token, datetime)`` pairs,
            denoting the timestamp when APNs has detected the device token is
            not available anymore, probably because application was
            uninstalled. You have to stop sending notifications to that device
            token unless it has been re-registered since reported timestamp.
            
            Unlike sending the message, you should fetch the feedback using
            non-cached connection. Once whole feedback has been read, this
            method will automatically close the connection.

            .. note::
                If the client fails to connect to APNs, probably because your
                network is down, then this method will raise the related
                exception. However, if connection is successfully established,
                but later on the IO fails, then this method will simply stop
                iterating. The rest of the failed tokens will be delivered
                during the next feedback session.

            Example::

                session = Session()
                # get non-cached connection, free from possible garbage
                con = session.new_connection("feedback_production", cert_string=db_certificate)
                service = APNs(con)
                try:
                    # on any IO failure after successfull connection this generator
                    # will simply stop iterating. you will pick the rest of the tokens
                    # during next feedback session.
                    for token, when in service.feedback():
                        # every time a devices sends you a token, you should store
                        # {token: given_token, last_update: datetime.datetime.now()}
                        last_update = get_last_update_of_token(token)

                        if last_update < when:
                            # the token wasn't updated after the failure has
                            # been reported, so the token is invalid and you should
                            # stop sending messages to it.
                            remove_token(token)
                except:
                    print "Check your network, I could not connect to APNs"

            :Returns:
                generator over ``(binary, datetime)``
        """
        # FIXME: this library is not idiot proof. If you store returned generator
        # somewhere, then yes, the connection will remain locked.
        for token, timestamp in self._connection.feedback():
            yield (token, self._datetime_from_timestamp(timestamp))

    # override if you use custom datetime or weird timezones
    def _datetime_from_timestamp(self, timestamp):
        """ Converts integer timestamp to ``datetime`` object. """
        return datetime.datetime.fromtimestamp(timestamp)


class Message(object):
    """ The notification message. """
    # JSON serialization parameters. Assume UTF-8 by default.
    json_parameters = {
        'separators': (',',':'),
        'ensure_ascii': False,
    }
    # Default expiry (1 day).
    DEFAULT_EXPIRY = datetime.timedelta(days=1)
    # Default message priority
    DEFAULT_PRIORITY = 10

    def __init__(self, tokens, alert=None, badge=None, sound=None, content_available=None,
                 expiry=None, payload=None, priority=DEFAULT_PRIORITY, extra=None,
                 **extra_kwargs):
        """ The push notification to one or more device tokens.

            Read more `about the payload
            <https://developer.apple.com/library/mac/documentation/NetworkingInternet/Conceptual/RemoteNotificationsPG/Chapters/ApplePushService.html#//apple_ref/doc/uid/TP40008194-CH100-SW1>`_.

            .. note::
                In order to stay future compatible this class doesn't transform
                provided arguments in any way. It is your responsibility to
                provide correct values and ensure the payload does not exceed
                the limit of 256 bytes. You can also generate whole payload
                yourself and provide it via ``payload`` argument. The payload
                will be parsed to init default fields like alert and badge.
                However if parsing fails, then these standard fields will
                become unavailable. If raw payload is provided, then other data
                fields like alert or sound are not allowed.

            :Arguments:
                - tokens (str or list): set of device tokens where to the message will be sent.
                - alert (str or dict): the message; read APNs manual for recognized dict keys.
                - badge (int or str): badge number over the application icon or special value such as "increment".
                - sound (str): sound file to play on arrival.
                - content_available (int): set to 1 to indicate new content is available.
                - expiry (int, datetime or timedelta): timestamp when message will expire.
                - payload (dict or str): JSON-compatible dictionary with the
                   complete message payload. If supplied, it is given instead
                   of all the other, more specific parameters.
                - priority (int): priority of the message, defaults to 10
                - extra (dict): extra payload key-value pairs.
                - extra_kwargs (kwargs): extra payload key-value paris, will be merged with ``extra``.
        """
        if payload is not None and ([v for v in (alert, badge, sound, content_available, extra) if v is not None] or extra_kwargs):
            # Raise an error if both `payload` and the more specific parameters are supplied.
            raise ValueError("Payload specified together with alert/badge/sound/content_available/extra.")

        # single token is provided, wrap as list
        if isinstance(tokens, six.string_types) or isinstance(tokens, six.binary_type):
            tokens = [tokens]

        self._tokens = tokens
        self._payload = payload
        self.priority = int(priority)  # has to be integer because will be formatted into a binary
        self.expiry = self._get_expiry_timestamp(expiry)

        if payload is not None and hasattr(payload, "get") and payload.get("aps"):
            # try to reinit fields from the payload
            aps = payload["aps"]
            self.alert = aps.get("alert")
            self.badge = aps.get("badge")
            self.sound = aps.get("sound")
            self.content_available = aps.get("content-available")
            self.extra = {k: v for k, v in six.iteritems(payload) if k != 'aps'}
        elif payload is None:
            # normal message initialization
            self.alert = alert
            self.badge = badge
            self.sound = sound
            self.content_available = content_available
            _extra = {}
            if extra:
                _extra.update(extra)
            if extra_kwargs:
                _extra.update(extra_kwargs)
            self.extra = _extra
            if 'aps' in self.extra:
                raise ValueError("Extra payload data may not contain 'aps' key.")
        # else: payload provided as unrecognized value, don't init fields,
        # they will raise AttributeError on access

    # override if you use funky expiry values
    def _get_expiry_timestamp(self, expiry):
        """ Convert expiry value to a timestamp (integer).
            Provided value can be a date or timedelta.
        """
        if expiry is None:
            # 0 means do not store messages at all. so we have to choose default
            # expiry, which is here 1 day.
            expiry = self.DEFAULT_EXPIRY

        if isinstance(expiry, datetime.timedelta):
            expiry = self._get_current_datetime() + expiry

        if isinstance(expiry, datetime.datetime):
            expiry = time.mktime(expiry.timetuple())

        return int(expiry)

    # override if you use funky timezones
    def _get_current_datetime(self):
        """ Returns current date and time. """
        return datetime.datetime.now()

    def __getstate__(self):
        """ Returns ``dict`` with ``__init__`` arguments.

            If you use ``pickle``, then simply pickle/unpickle the message object.
            If you use something else, like JSON, then::
                
                # obtain state dict from message
                state = message.__getstate__()
                # send/store the state
                # recover state and restore message
                message_copy = Message(**state)

            .. note::
                The message keeps ``expiry`` internally as a timestamp
                (integer).  So, if values of all other arguments are JSON
                serializable, then the returned state must be JSON
                serializable.  If you get ``TypeError`` when you instantiate
                ``Message`` from JSON recovered state, then make sure the keys
                are ``str``, not ``unicode``.

            :Returns:
                `kwargs` for `Message` constructor.
        """
        if self._payload is not None:
            return {
                'tokens': self.tokens,
                'expiry': self.expiry,
                'payload': self._payload,
                'priority': self.priority,
            }

        return {key: getattr(self, key) for key in ('tokens', 'alert', 'badge',
                    'sound', 'content_available', 'expiry', 'priority', 'extra')}
    
    def __setstate__(self, state):
        """ Overwrite message state with given kwargs. """
        self._tokens = state['tokens']
        self.extra = {}
        self.expiry = state['expiry']
        self.priority = state['priority']

        if 'payload' in state:
            self._payload = state['payload']
            if hasattr(self._payload, "get") and self._payload.get("aps"):
                aps = self._payload["aps"]
                self.alert = aps.get("alert")
                self.badge = aps.get("badge")
                self.sound = aps.get("sound")
                self.content_available = aps.get("content-available")
                self.extra = {k: v for k, v in six.iteritems(self._payload) if k != 'aps'}
        else:
            self._payload = None
            for key, val in six.iteritems(state):
                if key in ('tokens', 'expiry', 'priority'): # already set
                    pass
                elif key in ('alert', 'badge', 'sound', 'content_available'):
                    setattr(self, key, state[key])
                elif key == 'extra':
                    self.extra.update(state[key])
                else:
                    # legacy serialized object
                    self.extra[key] = val

    @property
    def tokens(self):
        """ List target device tokens. """
        return self._tokens

    @property
    def payload(self):
        """ Returns the payload content as a dict or raw ``payload`` argument value. """
        if self._payload is not None:
            return self._payload
        
        # in v.2 protocol no keys are required, but usually you specify
        # alert or content-available.
        aps = {}

        if self.alert is not None:
            aps['alert'] = self.alert

        if self.badge is not None:
            aps['badge'] = self.badge

        if self.sound is not None:
            aps['sound'] = self.sound

        if self.content_available is not None:
            aps['content-available'] = self.content_available

        ret = {
            'aps': aps,
        }
        
        if self.extra:
            ret.update(self.extra)

        return ret

    def get_json_payload(self):
        """ Convert message to JSON payload, acceptable by APNs. Must return byte string. """
        payload = self.payload
        if not isinstance(payload, six.string_types) and not isinstance(payload, six.binary_type):
            payload = json.dumps(payload, **self.json_parameters)

        # in python2 json will output utf-8 encoded str. in python3 json will output
        # a unicode string. So only for python3 in case of unicode string - encode.
        if not isinstance(payload, six.binary_type):
            payload = payload.encode("utf-8")

        return payload

    def batch(self, packet_size):
        """ Returns binary serializer. """
        payload = self.get_json_payload()
        assert isinstance(payload, six.binary_type), "Payload must be bytes/binary"
        return Batch(self._tokens, payload, self.expiry, self.priority, packet_size)

    def retry(self, failed_index, include_failed):
        """ Create new retry message with tokens from failed index. """
        if not include_failed:
            failed_index += 1

        failed = self._tokens[failed_index:]
        if not failed:
            # nothing to retry
            return None

        state = self.__getstate__()
        state['tokens'] = failed
        return Message(**state)


class Batch(object):
    """ Binary stream serializer. """
    # Frame version. Do not change unless you update binary formats too.
    VERSION = 2

    def __init__(self, tokens, payload, expiry, priority, packet_size):
        """ New serializer.

            :Arguments:
                - tokens (list): list of target target device tokens.
                - payload (str): JSON payload.
                - expiry (int): expiry timestamp.
                - priority (int): message priority.
                - packet_size (int): minimum chunk size in bytes.
        """
        self.tokens = tokens
        self.payload = payload
        self.expiry = expiry
        self.priority = priority
        self.packet_size = packet_size
        
    def __iter__(self):
        """ Iterate over serialized chunks. """
        messages = []
        buf = 0
        sent = 0

        # for all registration ids
        for idx, token in enumerate(self.tokens):
            tok = binascii.unhexlify(token)
            # |COMMAND|FRAME-LEN|{token}|{payload}|{id:4}|{expiry:4}|{priority:1}
            frame_len = 3*5 + len(tok) + len(self.payload) + 4 + 4 + 1 # 5 items, each 3 bytes prefix, then each item length
            fmt = ">BIBH{0}sBH{1}sBHIBHIBHB".format(len(tok), len(self.payload))
            message = pack(fmt, self.VERSION, frame_len,
                    1, len(tok), tok,
                    2, len(self.payload), self.payload,
                    3, 4, idx,
                    4, 4, self.expiry,
                    5, 1, self.priority)

            messages.append(message)
            buf += len(message)
            if buf >= self.packet_size:
                chunk = six.b("").join(messages)
                buf = 0
                prev_sent = sent
                sent += len(messages)
                messages = []
                yield prev_sent, chunk

        # last small chunk
        if messages:
            yield sent, six.b("").join(messages)


class Result(object):
    """ Result of send operation. """
    # all rerror codes {code: (explanation, can retry?, include failed token?)}
    ERROR_CODES = {
        1: ('Processing error', True, True),
        2: ('Missing device token', True, False), # looks like token was empty?
        3: ('Missing topic', False, True), # topic is encoded in the certificate, looks like certificate is wrong. bail out.
        4: ('Missing payload', False, True), # bail out, our message looks like empty
        5: ('Invalid token size', True, False), # current token has wrong size, skip it and retry
        6: ('Invalid topic size', False, True), # can not happen, we do not send topic, it is part of certificate. bail out.
        7: ('Invalid payload size', False, True), # our payload is probably too big. bail out.
        8: ('Invalid token', True, False), # our device token is broken, skipt it and retry
        10: ('Shutdown', True, False), # server went into maintenance mode. reported token is the last success, skip it and retry.
        None: ('Unknown', True, True), # unknown error, for sure we try again, but user should limit number of retries
    }

    def __init__(self, message, failure=None):
        """ Result of send operation. """
        self.message = message
        self._retry_message = None
        self._failed = {}
        self._errors = []

        if failure is not None:
            reason, failed_index = failure
            if reason not in self.ERROR_CODES:
                # one of "unknown" error codes
                reason = None

            expl, can_retry, include_failed = self.ERROR_CODES[reason]
            if can_retry:
                # may be None if failed on last token, which is skipped
                self._retry_message = message.retry(failed_index, include_failed)

            if reason == 10:
                # the Shutdown reason is not really an error, it just indicates
                # the server went into a maintenance mode and connection was closed.
                # The reported token is the last one successfully sent.
                pass
            elif not include_failed: # report broken token, it was skipped
                self._failed = {
                    message.tokens[failed_index]: (reason, expl)
                }
            else: # errors not related to broken token, global shit happened
                self._errors = [
                    (reason, expl)
                ]

            if LOG.isEnabledFor(logging.DEBUG):
                LOG.debug("Batch of %d tokens failed: %s.%s%s",
                          len(message.tokens), expl,
                          ' With errors.' if self._errors else '',
                          ' With failed tokens.' if self._failed else '')

    @property
    def errors(self):
        """ Returns list of ``(reason, explanation)`` pairs denoting severe errors,
            not related to failed tokens. The reason is an integer code as
            described in APNs tutorial.

            The following codes are considered to be errors:
                - ``(1, "Processing error")``
                - ``(3, "Missing topic")``
                - ``(4, "Missing payload")``
                - ``(6, "Invalid topic size")``
                - ``(7, "Invalid payload size")``
                - ``(None, "Unknown")``, usually some kind of IO failure.
        """
        return self._errors

    @property
    def failed(self):
        """ Reports failed tokens as ``{token : (reason, explanation)}`` mapping.

            Current APNs protocols bails out on first failed device token, so
            the returned dict will contain at most 1 entry. Future extensions
            may upgrade to multiple failures in a batch. The reason is the
            integer code as described in APNs tutorial.

            The following codes are considered to be token failures:
                - ``(2, "Missing device token")``
                - ``(5, "Invalid token size")``
                - ``(8, "Invalid token")``
        """
        return self._failed

    def needs_retry(self):
        """
            Returns True if there are tokens that should be retried.

            .. note::
                In most cases if ``needs_retry`` is true, then the reason of
                incomplete batch is to be found in ``errors`` and ``failed``
                properties. However, Apple added recently a special code *10
                - Shutdown*, which indicates server went into a maintenance
                mode before the batch completed. This response is not really an
                error, so the before mentioned properties will be empty, while
                ``needs_retry`` will be true.
        """
        return self._retry_message is not None

    def retry(self):
        """ Returns :class:`Message` with device tokens that can be retried.
       
            Current APNs protocol bails out on first failure, so any device
            token after the failure should be retried. If failure was related
            to the token, then it will appear in :attr:`failed` set and will be
            in most cases skipped by the retry message.
        """
        return self._retry_message



# Copyright 2014 Sardar Yumatov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import datetime

# python 3 support
import six


class BaseBackend(object):
    """ Interface definition for IO backends. """
    #: Default pool size.
    DEFAULT_POOL_SIZE = 5
    #: True if this backend can detect whether the connection is closed when
    # serving ``get_cached_connection()`` call. If False, then APNs will re-get
    # the connection if first IO operation fails. For example, POSIX sockets
    # can't detect if connection is closed until first read or write operation.
    can_detect_close = False

    def __init__(self, pool_size=DEFAULT_POOL_SIZE, use_cache_for_reconnects=True):
        """ Create new backend. The backend will keep a pool of connections for
            each ``(address, certificate)`` pair.

            Every new connetion will be eventually released. If it is not
            closed and the pool size is smaller than ``pool_size``, then
            connection will be added back to the pool. If ``pool_size`` is set
            to None, then pool will grow unlimited. If ``pool_size`` is 0, then
            pool will be effectivly disabled. The connection will be dropped
            from the pool if it gets closed.

            If ``can_detect_close`` is False, then session will try to
            reconnect if first IO operation fails on current connection, which
            is obtained from the cached pool. The argument
            ``use_cache_for_reconnects`` specifies how to obtain new connection
            when reconnecting. If set True, then new connection will be taken
            from the cached pool, otherwise the backend will always open a new
            connection.

            :Arguments:
                - pool_size (int): optimal pool size, None for unlimited size.
                - use_cache_for_reconnects (bool): True to use cache for re-connects.
        """
        self.pool_size = pool_size
        self.use_cache_for_reconnects = use_cache_for_reconnects
        self._connections = {} # (address, certificate) -> [connection]
        self._lock = self.create_lock()

    def get_certificate(self, cert_params):
        """ Create/load certificate from parameters. """
        raise NotImplementedError

    def create_lock(self):
        """ Provides semaphore with ``threading.Lock`` interface. """
        raise NotImplementedError

    def get_cached_connection(self, address, certificate, timeout=None):
        """ Obtain connection from the pool. Opens new connection if no free
            connection is found.

            :Arguments:
                - address (tuple): target (host, port).
                - certificate (:class:`Certificate`): certificate instance.
                - timeout (float): connection timeout in seconds
        """
        key = (address, certificate)
        try:
            self._lock.acquire()
            pool = self._connections.get(key)
            while pool:
                con = pool.pop(0)
                if not con.closed():
                    con.touch()
                    return con
        finally:
            self._lock.release()

        return self.get_new_connection(address, certificate, timeout=timeout)

    def release(self, connection):
        """ Release connection. Method stores connection in the pool if pool
            does not exceed the optimal size.

            :Arguments:
                - connection (object): connection, obtained from this backend.
        """
        if not connection.closed():
            if self.pool_size is None or self.pool_size > 0:
                try:
                    self._lock.acquire()
                    key = (connection.address, connection.certificate)
                    pool = self._connections.setdefault(key, [])
                    if self.pool_size is None or len(pool) < self.pool_size:
                        connection.touch()
                        pool.append(connection)
                        return
                finally:
                    self._lock.release()

            # pool is larger than the optimal size, close surplus connection
            connection.close()

    def get_new_connection(self, address, certificate, timeout=None):
        """ Open a new connection.
        
            :Arguments:
                - address (tuple): target (host, port).
                - certificate (:class:`Certificate`): certificate instance.
                - timeout (float): connection timeout in seconds
        """
        raise NotImplementedError

    def outdate(self, delta):
        """ Close open connections in the pool that are not used in more than
            ``delta`` time.

            You may call this method in a separate thread or run it in some
            periodic task. If you don't, then all connections will remain open
            until session is shut down. It might be an issue if you care about
            your open server connections.

            .. note::
                This method doesn't touch connetions currently being in use,
                even if you specify zero delta. It affects only unused
                connections in the pool.

            :Arguments:
                delta (datetime.timedelta): maximum age of unused connection.
        """
        try:
            self._lock.acquire()
            for key, pool in list(six.iteritems(self._connections)):
                new_pool = []
                for con in pool:
                    if not con.closed():
                        if not con.is_outdated(delta):
                            new_pool.append(con)
                        else:
                            con.close()

                if new_pool:
                    self._connections[key] = new_pool
                else:
                    del self._connections[key]
        finally:
            self._lock.release()

    def __del__(self):
        """ Close conections on destruction. """
        # NOTE: global package datetime can become None if session is stored in
        # a global variable and being garbage collected with the rest of the module.
        if datetime is not None:
            self.outdate(datetime.timedelta())


class BaseConnection(object):
    """ Connection interface. """

    def __init__(self, address, certificate):
        """ Open new connection to given address using given certificate. """
        self.address = address
        self.certificate = certificate
        self.touch()

    def touch(self):
        """ Reset last use timestamp. """
        self.last_use = datetime.datetime.now()

    def is_outdated(self, delta):
        """ Returns True if ``delta`` time has not been passed since the last use.
        
            :Arguments:
                - delta (datetime.timedelta): last use TTL.
        """
        return (datetime.datetime.now() - self.last_use) > delta

    def closed(self):
        """ Returns True if connection is closed via explicit :func:`close' call.
            If ``backend.can_detect_close`` is False, then this method is allowed
            to return False even if underlying connection has been closed by itself.

            .. warning::
                This method is not allowed to throw any exceptions, except
                those that indicate catastrophic events (such as
                out-of-memory). On any IO related failure this method should
                always return True.

            :Returns:
                True if connection is closed or check has been failed, otherwise
                returns False.
        """
        raise NotImplementedError

    def close(self):
        """ Close connection and free underlying resources. Does nothing if
            connection is already closed.
            
            .. warning::
                This method is not allowed to throw any exceptions, except
                those that indicate catastrophic events (such as
                out-of-memory). On any IO related failure this method should
                simply supress the problem and set :func:`closed` state True.
        """
        raise NotImplementedError

    def reset(self):
        """ Clear read and write buffers. Called before starting a new IO
            session using a cached connection. Prevents stale date from being
            read during a new session.
        """
        raise NotImplementedError

    def write(self, data, timeout):
        """ Write chunk of data. Either writes the whole sequence or fails with
            an IO related exception. The exception should be raised if
            connection is closed or not the whole data can be sent within given
            timeout or any other IO error occurs.

            :Arguments:
                - data (bytes): data to write.
                - timeout (float): IO timeout for write operation, None for unlimited blocking.
        """
        raise NotImplementedError

    def peek(self, size):
        """ Read chunk of data from the read buffer without blocking. Returns
            not empty sequence of bytes or None if read buffer is empty. This
            method is used to quickly look for a status frame while sending the
            message block by block. If connection is closed, then method should
            return None.

            :Arguments:
                - size (int): max number of bytes to read.

            :Returns:
                Not empty sequence of bytes or None if read buffer is empty or
                connection is closed.
        """
        raise NotImplementedError

    def read(self, size, timeout):
        """ Read chunk of data. Returns not empty sequence of bytes while
            blocking for at most timeout of seconds. Returns None if connection
            is closed. Raises exception if not a single byte can be read within
            given timeout or on any IO related failure.

            :Arguments:
                - size (int): max number of bytes to read.
                - timeout (float): IO timeout for read operation, None if unlimited.

            :Returns:
                Not empty sequence of bytes or None if connection has been closed.
        """
        raise NotImplementedError

    def __del__(self):
        """ Close conection on destruction. """
        self.close()
